import asyncio
from oasysdb.prelude import Record, Collection, Config, Database


NAME = "vectors"  # Initial collection name.
DIMENSION = 128
LEN = 100


def create_test_database() -> Database:
    """Creates a new test database with an initial collection."""

    db = Database.new("data/py")
    assert db.is_empty()

    # Create a test collection with random records.
    records = Record.many_random(dimension=DIMENSION, len=LEN)
    config = Config.create_default()
    collection = Collection.from_records(config, records)

    # Save the collection to the database.
    db.save_collection(name=NAME, collection=collection)
    assert not db.is_empty()

    return db


def test_open():
    db = Database(path="data/mt")
    assert db.is_empty()


def test_new():
    db = create_test_database()
    assert not db.is_empty()
    assert db.len() == 1


def test_get_collection():
    db = create_test_database()
    collection = db.get_collection(name=NAME)
    assert collection.len() == LEN


def test_save_collection():
    db = create_test_database()

    # Create a new collection and save it to the database.
    config = Config.create_default()
    collection = Collection(config=config)
    db.save_collection(name="test", collection=collection)

    assert db.len() == 2


def test_delete_collection():
    db = create_test_database()
    db.delete_collection(name=NAME)
    assert db.is_empty()


def test_flush():
    db = create_test_database()
    assert db.flush() > 0


def test_async_flush():
    db = create_test_database()
    assert asyncio.run(db.async_flush()) > 0



# flake8: noqa F821

from typing import Any, List, Dict
from oasysdb.vector import Vector, VectorID


class Config:
    """The configuration for the vector collection.

    Args:
    - ef_construction: Nodes to consider during index construction.
    - ef_search: Nodes to consider during the search.
    - ml: Layer multiplier of the HNSW index.
    - distance: Distance metric function.

    Distance metrics:
    - euclidean
    - dot
    - cosine
    - norm_cosine
    """

    ef_construction: int
    ef_search: int
    ml: float
    distance: str

    def __init__(
        self,
        ef_construction: int,
        ef_search: int,
        ml: float,
        distance: str
    ) -> None: ...

    @staticmethod
    def create_default() -> Config:
        """Returns a default configuration.

        Default values:
        - ef_construction: 40
        - ef_search: 15
        - ml: 0.2885
        - distance: euclidean
        """

    @staticmethod
    def default() -> Config:
        """Returns a default configuration object.
        This is an alias of create_default method and
        shared the same implementation.
        """


class Record:
    """The vector record to store in the collection.

    Args:
    - vector: Vector embedding of float values.
    - data: Metadata of the vector.

    Metadata types:
    - String
    - Number
    - List of metadata types
    - Dictionary of metadata types
    """

    vector: Vector
    data: Any

    def __init__(self, vector: List[float], data: Any) -> None: ...

    @staticmethod
    def random(dimension: int) -> Record:
        """Generates a random record with the given dimension
        with a random integer metadata.

        Args:
        - dimension: Vector dimension.
        """

    @staticmethod
    def many_random(dimension: int, len: int) -> List[Record]:
        """Generates a list of random records.

        Args:
        - dimension: Vector dimension.
        - len: Number of records.
        """


class Collection:
    """The collection of vectors and their metadata."""

    config: Config
    dimension: int
    relevancy: float

    def __init__(self, config: Config) -> None: ...

    @staticmethod
    def from_records(config: Config, records: List[Record]) -> Collection:
        """Build a collection from the given records.

        Args:
        - config: Collection configuration.
        - records: Records used to build the collection.
        """

    @staticmethod
    def build(config: Config, records: List[Record]) -> Collection:
        """Build a collection from the records.
        This is an alias of from_records method and shared
        the same implementation.

        Args:
        - config: Collection configuration.
        - records: Records used to build the collection.
        """

    def insert(self, record: Record) -> VectorID:
        """Inserts a record into the collection.

        Args:
        - record: Record to insert.
        """

    def insert_many(self, records: List[Record]) -> List[VectorID]:
        """Inserts multiple records into the collection.

        Args:
        - records: Records to insert.
        """

    def delete(self, id: VectorID) -> None:
        """Deletes a record from the collection.

        Args:
        - id: Vector ID to delete.
        """

    def get(self, id: VectorID) -> Record:
        """Returns a record from the collection.

        Args:
        - id: Vector ID to fetch.
        """

    def list(self) -> Dict[VectorID, Record]:
        """Returns a dictionary of records in the collection
        in the format of { VectorID: Record }.
        """

    def update(self, id: VectorID, record: Record) -> None:
        """Updates a record in the collection.

        Args:
        - id: Vector ID to update.
        - record: New record.
        """

    def search(self, vector: Vector, n: int) -> List[SearchResult]:
        """Searches for the nearest neighbors to
        the given vector using HNSW indexing algorithm

        Args:
        - vector: Vector to search.
        - n: Number of neighbors to return.
        """

    def true_search(self, vector: Vector, n: int) -> List[SearchResult]:
        """Searches for the nearest neighbors using brute force.

        Args:
        - vector: Vector to search.
        - n: Number of neighbors to return.
        """

    def len(self) -> int:
        """Returns the number of records in the collection."""

    def is_empty(self) -> bool:
        """Returns True if the collection is empty."""

    def contains(self, id: VectorID) -> bool:
        """Returns True if the vector ID is in the collection."""


class SearchResult:
    """The result of a search operation on the collection."""

    id: int
    distance: float
    data: Any



# flake8: noqa F821

from typing import List


class Vector:
    """The vector embedding containing float values."""

    def __init__(self, vector: List[float]) -> None: ...

    def len(self) -> int:
        """Returns the length of the vector."""

    def is_empty(self) -> bool:
        """Returns True if the vector is empty."""

    def to_list(self) -> List[float]:
        """Returns the vector as a list of floats."""

    @staticmethod
    def random(dimension: int) -> Vector:
        """Generates a random vector of the given dimension."""


class VectorID:
    """Identifier of the vector in the collection."""

    def __init__(self, id: int) -> None: ...

    def is_valid(self) -> bool:
        """Returns True if the vector ID is valid."""



from oasysdb.prelude import Config, Record, Collection, Vector, VectorID

DIMENSION = 128
LEN = 100


def create_test_collection() -> Collection:
    """Creates a collection with random records for testing."""
    records = Record.many_random(dimension=DIMENSION, len=LEN)
    config = Config.default()
    collection = Collection.from_records(config=config, records=records)

    assert collection.len() == len(records)
    return collection


def test_create_config():
    default = Config.create_default()

    # Create config based on the default.
    config = Config(
        ef_construction=40,
        ef_search=15,
        ml=0.2885,
        distance="euclidean"
    )

    assert config.ef_construction == default.ef_construction
    assert config.ef_search == default.ef_search
    assert config.ml == default.ml
    assert config.distance == default.distance


def test_create_record():
    vector = [0.1, 0.2, 0.3]
    data = {"text": "This is an example."}
    record = Record(vector=vector, data=data)

    assert len(record.vector) == len(vector)
    assert record.data == data


def test_generate_random_record():
    record = Record.random(dimension=DIMENSION)
    assert len(record.vector) == DIMENSION
    assert isinstance(record.data, int)


def test_generate_many_random_records():
    records = Record.many_random(dimension=DIMENSION, len=LEN)
    assert len(records) == LEN


def test_create_collection():
    config = Config.create_default()
    collection = Collection(config=config)

    assert collection.config.ml == config.ml
    assert collection.is_empty()


def test_build_collection():
    collection = create_test_collection()
    assert collection.contains(VectorID(0))
    assert not collection.is_empty()


def test_insert_record():
    collection = create_test_collection()
    record = Record.random(dimension=128)
    collection.insert(record)

    assert collection.len() == LEN + 1
    assert collection.contains(VectorID(LEN))


def test_insert_record_invalid_dimension():
    collection = create_test_collection()
    record = Record.random(dimension=100)

    # Insert should raise an exception because the
    # vector dimension is invalid.
    try:
        collection.insert(record)
        assert False
    except Exception as e:
        assert "invalid vector dimension" in str(e).lower()

    assert collection.len() == LEN


def test_insert_many_records():
    collection = create_test_collection()
    records = Record.many_random(dimension=DIMENSION, len=LEN)
    collection.insert_many(records)

    assert collection.len() == 2 * LEN
    assert all(collection.contains(VectorID(i)) for i in range(LEN, 2 * LEN))


def test_delete_record():
    collection = create_test_collection()

    id = VectorID(0)
    collection.delete(id)

    assert not collection.contains(id)
    assert collection.len() == LEN - 1


def test_get_record():
    collection = create_test_collection()

    id = VectorID(0)
    record = collection.get(id)

    assert record is not None
    assert record.data is not None


def test_update_record():
    collection = create_test_collection()

    id = VectorID(0)
    record = Record.random(dimension=128)
    collection.update(id, record)

    assert collection.contains(id)
    assert collection.get(id).data == record.data


def test_search_record():
    collection = create_test_collection()
    collection.relevancy = 4.5

    vector = Vector.random(dimension=DIMENSION)
    n = 10

    # Search for approximate neighbors and true neighbors.
    results = collection.search(vector, n=n)
    true_results = collection.true_search(vector, n=n)

    assert len(results) == n
    assert len(true_results) == n

    # Make sure the first result of the approximate search
    # is somewhere in the true results.
    assert results[0].id in [true.id for true in true_results]

    # Check if the result distances are within the relevancy.
    assert results[-1].distance <= collection.relevancy
    assert true_results[-1].distance <= collection.relevancy


def test_set_dimension():
    config = Config.create_default()
    collection = Collection(config=config)

    # Set the collection dimension to 100.
    collection.dimension = 100

    # When inserting a record with a different dimension,
    # the collection should raise an exception.
    try:
        record = Record.random(dimension=128)
        collection.insert(record)
        assert False
    except Exception as e:
        assert "invalid vector dimension" in str(e).lower()


def test_list_records():
    collection = create_test_collection()
    records = collection.list()

    assert len(records) == collection.len()
    assert all(isinstance(k, VectorID) for k in records.keys())
    assert all(isinstance(v, Record) for v in records.values())


def test_collection_distance_euclidean():
    config = Config.default()
    collection = Collection(config=config)

    # Insert records.
    k = 5
    records = Record.many_random(dimension=DIMENSION, len=k)
    collection.insert_many(records)

    # Search for the record.
    query = Vector.random(dimension=DIMENSION)
    results = collection.search(query, n=k)

    # Sort result based on distance ascending.
    sort = sorted(results, key=lambda x: x.distance)

    for i in range(k):
        assert results[i].distance == sort[i].distance


def test_collection_distance_cosine():
    config = Config.create_default()
    config.distance = "cosine"
    collection = Collection(config=config)

    # Insert records.
    k = 5
    records = Record.many_random(dimension=DIMENSION, len=k)
    collection.insert_many(records)

    # Search for the record.
    query = Vector.random(dimension=DIMENSION)
    results = collection.search(query, n=k)
    true_results = collection.true_search(query, n=k)

    # Sort result based on distance descending.
    sort = sorted(results, key=lambda x: x.distance, reverse=True)

    for i in range(k):
        assert results[i].distance == true_results[i].distance
        assert results[i].distance == sort[i].distance
